#ifdef USE_DOUBLES
typedef double float_type;
#else
typedef float float_type;
#endif

#include <string>
#include <vector>

class CFTree{

    //***************************Node class*****************************
private:
    class Node{
    public:
        int n; //number of points that made up the cluster represented by the node
        float_type *linsum;
        float_type squaredsum;
        int numchildren; //number of subclusters
        bool isleaf;
        Node *parent;
        Node **children;
        int dim;
        int branchingfactor;
        float_type *centroid;
        std::vector<int> assigned_points;

        Node(bool isleaf, int dim, int branchingfactor);
        ~Node();
        void calculate_centroid();
        const float_type *get_centroid(); //pointer to constant data -> can only access the centroid, but not change it from outside
        bool append_child(Node *node);
        bool remove_child(Node *node);
        void print(std::string indent, bool last);
        void print_self();
    };
    //********************************************************

private:
    float_type threshold;
    int wavelet_length; //total length of each wavelet in the data
    int used_resolution; //number of values in a wavelet that should be used for creating the CF tree
    int wavelets_per_point; //number of timelines per data point
    int branchingfactor; //maximum number of children a Node may have (= first-level subclusters)
    int num_points; //number of data points that were inserted into the tree
    Node *root; //pointer to the root of the tree
    std::vector<Node*> leaves; //vector of pointers to all the leaves in the tree (for getting the clustering results)

    //tree building methods
    Node *find_leaf(const float_type *point);
    void modify_leaf(const float_type *point, Node *leaf, int point_id);
    void insert_sibling_node(Node *node, Node *newnode);
    void redistribute_nodes(Node *parent, Node *newparent, Node *newnode, Node *a, Node *b);
    void find_farthest_pair(Node *parent, Node *newnode, Node **a, Node **b);

    //utility methods
    void set_as_linsum(Node *node, const float_type *point);
    void add_to_linsum(Node *node, const float_type *point);
    float_type calculate_squaredsum(const float_type *point);
    float_type compute_diameter(Node *node, const float_type *point);
    float_type dist_between_nodes(Node *node1, Node *node2); //distance between two centroids
    float_type dist_to_centroid(const float_type *point, const float_type *centroid); //distance point-to-centroid

public:
    CFTree(float_type threshold, int wavelet_length, int used_resolution, int wavelets_per_point, int branchingfactor);
    ~CFTree();
    void insert(const float_type *data, int num_points);
    void print(); //print the whole tree
    void print_leaves(); //print the leaf nodes that are stored in leaves vector
    bool get_clustering_results(int *k, int *&cluster_assignments, float_type *&centroids, int *&points_per_cluster);

};
