#include <limits>
#include <iostream>
#include <math.h>
#include <cassert>

#include "ikmeans.h"


//Constructor
IKmeans::IKmeans(int k, int max_iterations,
    float_type *data, int num_points, int wavelets_per_point, int wavelet_length,
    int *cluster_assignments, float_type *centroids, int *points_per_cluster,
    int *coeffs_lengths, int num_levels, int current_level
):k(k), max_iterations(max_iterations), data(data), num_points(num_points), wavelets_per_point(wavelets_per_point), wavelet_length(wavelet_length),
cluster_assignments(cluster_assignments), centroids(centroids), points_per_cluster(points_per_cluster),
coeffs_lengths(coeffs_lengths), num_levels(num_levels), current_level(current_level), point_offset(wavelets_per_point * wavelet_length)
{}




void IKmeans::recalculate_centroids(int resolution){

    #ifdef VERBOSE
    std::cout << "Recalculating centroids at resolution " << resolution << std::endl;
    #endif

    //for each centroid: set all values to zero
    int centroids_length = k * wavelet_length * wavelets_per_point;
    for (int i = 0; i < centroids_length; ++i){
        centroids[i] = 0;
    }

    //for each point, add the appropriate values to the centroid of the cluster the point has been assigned to
    int assigned_cluster, n;
    for (int point_id = 0; point_id < num_points; ++point_id){
        assigned_cluster = cluster_assignments[point_id];
        n = points_per_cluster[assigned_cluster];
        for (int i = 0; i < wavelets_per_point; ++i){
            for (int j = 0; j < resolution; ++j){
                centroids[assigned_cluster*point_offset + i*wavelet_length + j] += data[point_id*point_offset + i*wavelet_length + j] / n;
            }
        }
    }//end point_id
}



int IKmeans::cluster(){

    //calculate resolution to start with
    int current_resolution = 0;
    for (int i = 0; i <= current_level; ++i)
        current_resolution += coeffs_lengths[i];

    #ifdef VERBOSE
    std::cout << "Incremental-K-means clustering, starting with a resolution of " << current_resolution << std::endl;
    #endif

    //utility variables
    bool converged;
    float_type *point = nullptr, *centroid = nullptr;
    float_type dist, current_dist;
    int assigned_cluster;


    int iteration = 1;
    for (; iteration <= max_iterations; ++iteration){
        #ifdef VERBOSE
        std::cout << "\nIteration " << iteration << std::endl;
        #endif

        //set up stuff for current iteration
        converged = true;
        recalculate_centroids(current_resolution); //recalculates all the centroids


        //cluster points
        for (int point_id = 0; point_id < num_points; ++point_id){ //for each point:

            point = data + point_id*point_offset;

            //find closest cluster
            dist = std::numeric_limits<float_type>::infinity();
            assigned_cluster = -1;
            for (int cluster_id = 0; cluster_id < k; ++cluster_id){ //for each centroid:
                centroid = centroids + cluster_id*point_offset;
                current_dist = distance(point, centroid, current_resolution);
                if (current_dist < dist){
                    dist = current_dist;
                    assigned_cluster = cluster_id;
                }
            }//end for each centroid

            //sanity check
            assert(assigned_cluster > -1);

            #ifdef VERBOSE
            std::cout << "Assigning point " << point_id << " to cluster " << assigned_cluster << " with distance " << dist << std::endl;
            #endif

            //check if the cluster assignment of the current point has changed
            if (assigned_cluster != cluster_assignments[point_id]){
                #ifdef VERBOSE
                std::cout << "Cluster assignment changed!" << std::endl;
                #endif

                points_per_cluster[cluster_assignments[point_id]]--;
                points_per_cluster[assigned_cluster]++;
                cluster_assignments[point_id] = assigned_cluster;
                converged = false;
            }
        } //end for each point


        //if no points changed cluster assignment, stop
        if (converged){
            #ifdef VERBOSE
            std::cout << "Reached convergence" << std::endl;
            #endif

            break;
        }

        //adjust resolution (if possible)
        if (current_level < num_levels-1){
            current_resolution += coeffs_lengths[++current_level];
        }
        #ifdef VERBOSE
        std::cout << "Resolution for next iteration is " << current_resolution << std::endl;
        #endif



    } //end for iteration

    #ifdef VERBOSE
    std::cout << "Stopped after " << iteration << " iterations" << std::endl;
    #endif


    return iteration;


}



float_type IKmeans::distance(const float_type *p1, const float_type *p2, int used_resolution){

    int point_data_length = wavelet_length * wavelets_per_point;
    float_type dist = 0;
    float_type help;
    for (int i = 0; i < point_data_length; i += wavelet_length){
        for (int j = 0; j < used_resolution; j++){
            help = p1[i+j] - p2[i+j];
            dist += help*help;
        }
    }
    return dist;
}
